/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include "client.h"
#include <stdio.h>
#include <string.h>

#ifdef CLIENT
# define FILEPREF	"client.pref"
#else
# define FILEPREF	"none.pref"
#endif

PREF_UL *prefs;

#define YES( c )	( c == 'y' || c == 'Y' )

/* for fields regarding other applications, like sound or video output, 0
if a numeric filed or '\0' if a string will set defaults */
/*** should detect errors with scanfs and etc */
int loadprefs ( void ) {
	FILE	*input;
	char	temp[512], *value;
	int	i = 0;
	long	t[3];
	
	if ( !(prefs = (PREF_UL *)malloc(sizeof(PREF_UL))) ) {
		fprintf( stderr, "Couldn't allocate prefs! ARGH!\n" );
		return FATALERROR;
	}
		

	prefs->useobjf = False;
#ifdef CLIENT
	prefs->kph = True;
	prefs->map = True;
	prefs->position = True;
#endif
	prefs->dltrack = False;

	prefs->updatefreq = 10;
	prefs->updatefreqts = timetots( 0, 1/prefs->updatefreq, 0 );
	prefs->maxupdates = 20;
	prefs->serverport = DEFPORT;
	strcpy( prefs->servername, "localhost" );

#ifdef CLIENT
	prefs->engprefs.aa = False;
	prefs->engprefs.gtype = FLAT;
	prefs->engprefs.fog.type = NONE;
	/* the rest of the fog doesn't belong here */

	prefs->maxsoundcars = 10;
	prefs->maxfrontviewcars = 10;
	prefs->maxrearviewcars = 5;

	prefs->controller = KEYBOARD;
	
	prefs->keybset.left = GIIK_Left;
	prefs->keybset.right = GIIK_Right;
	prefs->keybset.accel = GIIK_Up;
	prefs->keybset.brake = GIIK_Down;
	prefs->keybset.gearup = 'a';
	prefs->keybset.geardown = 'z';
	prefs->keybset.geardown = ' ';
	prefs->keybset.changeview = 'v';
	
	prefs->height = 0;
	prefs->width = 0;
	prefs->depth = 0;
	prefs->single = False;
	prefs->usedb = True;

	prefs->frequency = 0;
	prefs->sounddevice[0] = '\0';
	
	prefs->maxmessages = 15;
	prefs->messagetime = timetots( 0, 6, 0 ); /* six seconds */
#endif
	
	sprintf( temp, "%s%s", ULDIR, FILEPREF );
	if ( (input = fopen( temp, "r" )) ) {
		while ( fgets( temp, sizeof( temp ), input )) {
			i++;
			
			if ( value = strchr( temp, '#' ) )
				*value = '\0';
			
			value = strchr( temp, '=' );
			if ( !value )
				continue;
			*value++ = '\0';

			while( isspace( *value ) )
				value++;

			if ( strstr( temp, "useobjf" ) )
				prefs->useobjf = ( YES(*value) ? True : False );
#ifdef CLIENT
			else if ( strstr( temp, "kph" ) )
				prefs->kph = ( YES(*value) ? True : False );
			else if ( strstr( temp, "map" ) )
				prefs->map = ( YES(*value) ? True : False );
			else if ( strstr( temp, "position" ) )
				prefs->position = ( YES(*value) ? True : False );
#endif
			else if ( strstr( temp, "dltrack" ) )
				prefs->dltrack = ( YES(*value) ? True : False );
			else if ( strstr( temp, "servername" ) )
				sscanf( value, "%[^\n]", prefs->servername );
			else if ( strstr( temp, "serverport" ) )
				prefs->serverport = atoi( value );
			else if ( strstr( temp, "updates per second" ) ) {
				prefs->updatefreq = atoi( value );
				prefs->updatefreqts = timetots( 0, 1/prefs->updatefreq, 0 );
			}
			else if ( strstr( temp, "max updates until precise update" ) )
				prefs->maxupdates = atoi( value );
#ifdef CLIENT
			else if ( strstr( temp, "antialiasing" ) )
				prefs->engprefs.aa = ( YES(*value) ? True : False );
			else if ( strstr( temp, "graphics" ) ) {
				if ( strstr( value, "VERTICES" ) )
					prefs->engprefs.gtype = VERTICES;
				else if ( strstr( value, "WIREFRAME" ) )
					prefs->engprefs.gtype = WIRE;
				else if ( strstr( value, "FLAT" ) )
					prefs->engprefs.gtype = FLAT;
				else if ( strstr( value, "TEXTURE" ) )
					prefs->engprefs.gtype = TEXTURE;
				else if ( strstr( value, "GOURAUD" ) )
					prefs->engprefs.gtype = GOURAUD;
				else if ( strstr( value, "TEXTUREGOURAUD" ) )
					prefs->engprefs.gtype = TEXTUREGOURAUD;
				else
					fprintf( stderr, "Unknown graphics type at %s, line %d\n",
							FILEPREF, i );
			}
			else if ( strstr( temp, "fogtype" ) ) {
				if ( strstr( value, "NONE" ) )
					prefs->engprefs.fog.type = NONE;
				else if ( strstr( value, "LINEAR" ) )
					prefs->engprefs.fog.type = LINEAR;
				else if ( strstr( value, "EXP" ) )
					prefs->engprefs.fog.type = EXP;
				else if ( strstr( value, "EXP2" ) )
					prefs->engprefs.fog.type = EXP2;
				else
					fprintf( stderr, "Unknown fog type at %s, line %d\n",
							FILEPREF, i );
			}
			else if ( strstr( temp, "maxsoundcars" ) )
				prefs->maxsoundcars = atoi( value );
			else if ( strstr( temp, "maxfrontviewcars" ) )
				prefs->maxfrontviewcars = atoi( value );
			else if ( strstr( temp, "maxrearviewcars" ) )
				prefs->maxrearviewcars = atoi( value );
			else if ( strstr( temp, "controller" ) ) {
				if ( strcmp( value, "joystick" ) )
					prefs->controller = JOYSTICK;
				else if ( strcmp( value, "keyboard" ) );
					prefs->controller = KEYBOARD;
			}
			else if ( strstr( temp, "height" ) )
				prefs->height = atoi( value );
			else if ( strstr( temp, "width" ) )
				prefs->width = atoi( value );
			else if ( strstr( temp, "depth" ) ) {
				prefs->depth = atoi( value );
				if ( prefs->depth != 32 && prefs->depth != 24 &&
				     prefs->depth != 16 && prefs->depth != 8 &&
				     prefs->depth != 0 ) {
					fprintf( stderr, "Depth value %d not valid. Try 32, 24, 16 or 8 or 0\n",
							prefs->depth );
					prefs->depth = 0;
				}
			}
			else if ( strstr( temp, "single" ) )
				prefs->single = ( YES(*value) ? True : False );
			else if ( strstr( temp, "usedb" ) )
				prefs->usedb = ( YES(*value) ? True : False );
			else if ( strstr( temp, "sound" ) )
				prefs->sound = ( YES(*value) ? True : False );
			else if ( strstr( temp, "frequency" ) )
				prefs->frequency = atoi( value );
			else if ( strstr( temp, "sound device" ) )
				sscanf( value, "%[^\n]", prefs->sounddevice );
			else if ( strstr( temp, "max messages" ) )
				prefs->maxmessages = atoi( value );
			else if ( strstr( temp, "message time" ) ) {
				sscanf( value, "%d:%d:%d", &t[0], &t[1], &t[2] );
				prefs->messagetime = timetots( t[0], t[1], t[2] );
			}
#endif
			else
				fprintf( stderr, "Syntax error in %s, line %d; ignoring\n",
					 FILEPREF, i );
		}
		fclose( input );
		return OK;
	}
	else {
		fprintf( stderr, 
			"Could not open preferences file: %s. Trying to create a new one\n",
			FILEPREF );
		saveprefs();
		return OK;
	}
		
}

#define YESNO( d )		( d == True ? "yes" : "no" )

void saveprefs ( void ) {
	FILE		*out;
	unsigned char	temp[512];

	sprintf( temp, "%s%s", ULDIR, FILEPREF );
	if ( (out = fopen( temp, "w" )) ) {
		fprintf( out, "# Unlimited Simulator client preferences file\n" );
		fprintf( out, "useobjf = %s\n", YESNO( prefs->useobjf ) );
#ifdef CLIENT
		fprintf( out, "kph = %s\n", YESNO( prefs->kph ) );
		fprintf( out, "map = %s\n", YESNO( prefs->map ) );
		fprintf( out, "position = %s\n", YESNO( prefs->position ) );
#endif
		fprintf( out, "dltrack = %s\n", YESNO( prefs->dltrack ) );
		fprintf( out, "servername = %s\n", prefs->servername );
		fprintf( out, "serverport = %d\n", prefs->serverport );
		fprintf( out, "updates per second = %d\n", prefs->updatefreq );
		fprintf( out, "max updates until precise update = %d\n", prefs->maxupdates );
#ifdef CLIENT
		fprintf( out, "antialiasing = %s\n", YESNO( prefs->engprefs.aa ) );
		switch ( prefs->engprefs.gtype ) {
			case VERTICES:
				fprintf( out, "graphics = VERTICES\n" );
				break;
			case WIRE:
				fprintf( out, "graphics = WIREFRAME\n" );
				break;
			case FLAT:
				fprintf( out, "graphics = FLAT\n" );
				break;
			case TEXTURE:
				fprintf( out, "graphics = TEXTURE\n" );
				break;
			case GOURAUD:
				fprintf( out, "graphics = GOURAUD\n" );
				break;
			case TEXTUREGOURAUD:
				fprintf( out, "graphics = TEXTUREGOURAUD\n" );
				break;
		}
		switch ( prefs->engprefs.fog.type ) {
			case NONE:
				fprintf( out, "fogtype = NONE\n" );
			case LINEAR:
				fprintf( out, "fogtype = LINEAR\n" );
			case EXP:
				fprintf( out, "fogtype = EXP\n" );
			case EXP2:
				fprintf( out, "fogtype = EXP2\n" );
		}
		fprintf( out, "maxsoundcars = %d\n", prefs->maxsoundcars );
		fprintf( out, "maxfrontviewcars = %d\n", prefs->maxfrontviewcars );
		fprintf( out, "maxrearviewcars = %d\n", prefs->maxrearviewcars );
		fprintf( out, "controller = ");
		switch ( prefs->controller ) {
			case JOYSTICK:
				fprintf( out, "%s\n", "joystick" );
				break;
			case KEYBOARD:
				fprintf( out, "%s\n", "keyboard" );
				break;
		}
		fprintf( out, "width = %d\n", prefs->width );
		fprintf( out, "height = %d\n", prefs->height );
		fprintf( out, "depth = %d\n", prefs->depth );
		fprintf( out, "single = %s\n", YESNO(prefs->single) );
		fprintf( out, "usedb = %s\n", YESNO(prefs->usedb) );
		fprintf( out, "sound = %s\n", YESNO(prefs->sound) );
		fprintf( out, "frequency = %d\n", prefs->frequency );
		fprintf( out, "max messages = %d\n", prefs->maxmessages );
		fprintf( out, "message time = %d:%d:%d\n", 
			(prefs->messagetime/1000)/60,
			(prefs->messagetime/1000)%60, prefs->messagetime%1000 );
		if ( prefs->sounddevice[0] != '\0' )
			fprintf( out, "sounddevice = %s\n", prefs->sounddevice );
#endif
		fprintf( out, "# End of configuration file" );
		fclose( out );
	}
	else
		fprintf( stderr, "Could not open preferences file to save\n" );
}
