/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/

/* This file is a well documented skeleton to create your own client.
All functions that are recognized by the main client are declared here,
and some useful code too. */

#include "client.h"
#include <stdio.h>
#include <string.h>
/* #define AI_ONLY */
#ifndef AI_ONLY
# include <ggi/keyboard.h>
#endif

/* if some special version is required, uncomment the line below and fulfill
it correctly: */
/* #define NEEDVERSION	"" */

/* this structure will contain all relevant data for your client. Feel free
to rename it to whatever you want. */
struct mycli {
	/* some data */
};
typedef struct mycli MYCLI;


/* this function initializes all data. Here you read the client's specs,
initialize functions, etc. This function is called once when the client is
loaded. Do not rename it. Note that CLIENT_UL *own is already allocated. */
int initClient ( void ) {
	MYCLI	*myc;
	
	/* check for UL version */
#ifdef NEEDVERSION
	if ( strcmp( NEEDVERSION, VERSION ) > 0 ) {
		fprintf( stderr, "This plugin needs Unlimited Simulator version %s or above",
				NEEDVERSION );
		return FATALERROR;
	}
#endif
	/* allocate memory */
	if ( !(myc = (MYCLI *)malloc(sizeof(MYCLI))) ) {
		fprintf( stderr, "Could not allocate memory for MYCLI at initClient" );
		return FATALERROR; /* if something bad happens, return a
					FATALERROR. UL will handle the rest
					and quit. */
	}
	own->struc = myc;

	/* allocate and initialize your own stuff */

	/* other miscellaneous initializations */

	return OK;
}

/* this function frees all data. It is called once when the simulation is
over, to clean up everything. Do not rename it. */
void closeClient ( void ) {
	/* free general stuff that was allocated BY YOU */

	
	/* here you free all fields of client->struc */


	/* freeing the structure */
	free( own->struc );
	
	/* the rest is done by the Unlimited Simulator itself. */
}

/* When some event is received, such as a key press, which is relevant to the
client it's passed to this function. If this is an AI only client (the NONE
type), uncomment the line "#define AI_ONLY" on the beginning of the file;
the input is then ignored.*/
int handleEvent ( EVENT_UL event, int label, int sym ) {
	switch ( event ) {
#ifndef AI_ONLY
	  case KEYPRESS_UL:
	  case KEYREPEAT_UL:
		switch ( label ) {
		/* labels are defined in the file (LIBDIR)/ggi/keyboard.h.
		   sym is the symbolic code, such as '+', 'a', etc. Some common
		   keys are set below. */
		        case GIIK_Up:
				break;
			case GIIK_Down:
				break;
		        case GIIK_Left:
				break;
		        case GIIK_Right:
				break;
	        	case GIIK_Enter:
				break;
			default:
				break;
		}
		break;
	
	  case KEYRELEASE_UL:
		switch ( label ) {
		/* labels are defined in the file (LIBDIR)/ggi/keyboard.h.
		   sym is the symbolic code, such as '+', 'a', etc. Some common
		   keys are set below. */
		        case GIIK_Up:
				break;
			case GIIK_Down:
				break;
		        case GIIK_Left:
				break;
		        case GIIK_Right:
				break;
	        	case GIIK_Enter:
				break;
			default:
				break;
		}
		break;
#endif /* AI_ONLY */
	  /* other events in future */
	}
	return OK; /* there's no reason to return ERROR or FATALERROR here,
	except if some malloc didn't work, or whatever. If just the key is
	unknown, return OK, otherwise the program will quit. */
} 


/* these functions will handle special messages sent by the network. If you
want to be able to use this feature, uncomment the "#define SPECIAL_MESSAGE"
line below. Remember that your special messages will only be understood by
other clients compatible to this one. */
/* #define SPECIAL_MESSAGE */
#ifdef SPECIAL_MESSAGE
int handleSpecialMessage ( CLIENTM_UL *request, SERVERM_UL *response ) {
	if ( response == NULL || request == NULL ) {
		/* something went wrong, but probably it's not your fault.
		  Ignore this when coding. */
		fprintf( stderr, "Error, request = %p, response = %p. Please contact the writer\n"
				"of the client you're using and/or a UL developer, transcribing this message.\n",
				request, response );
		return FATALERROR;
	}

	/* parse response->special.field and fill request->special.data, 
	  if needed. UL will take care of sending it. The request->special.to
	  field in CLIENTM_UL structure or the request->special.from field
	  in SERVERM_UL are explained here:
	  positive: to the client with that number (SIMPLECLIENT_UL.id>>16)
	  0: it will be sent to all clients
	  negative: to all except the one whose number is the opposite of
	    the number given. */
	
	switch ( response->special.type ) { /* this is the message type, which
						is defined by you */
	}

	/* now UL will send it. Add some condition if there may be no replies. */
	sendSpecialMessageR( request );
	return OK;
}

void sendSpecialMessage ( int to, unsigned char type, unsigned char *data ) {
	CLIENTM_UL	req;
	
	req->request = UL_SPECIAL;
	req->special.from = own->number>>16;
	req->special.to = to;
	req->special.type = type;
	req->special.data = data;
	sendSpecialMessageR( &req );
}
#endif

/* this is the main loop. It's called every loop of the client main program,
and here you should do: physics simulation, AI, etc, etc. Note that, although
mainLoop() and handleEvent() are different functions, you can place events
in a stack and handle they here. */
int mainLoop ( void ) {
	/* Welcome to the core... */
	
	return OK;
}
