/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include "mathlib.h"

static MATRIX_UL prm;
static VECTOR_UL prv;

static const MATRIX_UL ident = { { 1.0, 0.0, 0.0, 0.0 },
				 { 0.0, 1.0, 0.0, 0.0 },
				 { 0.0, 0.0, 1.0, 0.0 },
				 { 0.0, 0.0, 0.0, 1.0 } };

void identity ( MATRIX_UL a ) {
	memcpy( a, ident, sizeof(double)<<4 );
}

void copyM ( MATRIX_UL target, MATRIX_UL src ) {
	memcpy( target, src, sizeof(double)<<4 );
}

void addm ( MATRIX_UL a, MATRIX_UL b, MATRIX_UL r ) {
	int i, j;
	for ( i = 0; i < 4; i++ )
		for ( j = 0; j < 4; j++ )
			r[i][j] = a[i][j] + b[i][j];
}

void subm ( MATRIX_UL a, MATRIX_UL b, MATRIX_UL r ) {
	int i, j;
	for ( i = 0; i < 4; i++ )
		for ( j = 0; j < 4; j++ )
			r[i][j] = a[i][j] - b[i][j];
}

void scalarmM ( MATRIX_UL a, const double m, MATRIX_UL r ) {
	int i, j;
	for ( i = 0; i < 4; i++ )
		for ( j = 0; j < 4; j++ )
			r[i][j] = a[i][j]*m;
}

void multM ( MATRIX_UL a, MATRIX_UL b, MATRIX_UL r ) {
	int i, j, t = 0;
	double *p;
	
	if ( &a[0][0] != &r[0][0] && &b[0][0] != &r[0][0] ) {
		p = r[0];
		t = 1;
	}
	else
		p = prm[0];
	
	for ( i = 0; i < 4; i++ )
		for ( j = 0; j < 4; j++, p++ )
			*p = a[i][0]*b[0][j] +
			     a[i][1]*b[1][j] +
			     a[i][2]*b[2][j] +
			     a[i][3]*b[3][j];
	if ( t == 0 )
		copyM( r, prm );
}

void multVM ( MATRIX_UL m, VECTOR_UL *v, VECTOR_UL *result ) {
	VECTOR_UL *p;
	if ( v == result )
		p = &prv; 
	else
		p = result;
	
	p->x = v->x*m[0][0] + v->y*m[1][0] + v->z*m[2][0] + m[3][0];
	p->y = v->x*m[0][1] + v->y*m[1][1] + v->z*m[2][1] + m[3][1];
	p->z = v->x*m[0][2] + v->y*m[1][2] + v->z*m[2][2] + m[3][2];
	
	if ( v == result )
		result->x = p->x;
		result->y = p->y;
		result->z = p->z;
}
