/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include <math.h>
#include "mathlib.h"

/* using 256 (0xFF) or 0xFFF (4080) or even the giant but precise to hell
0xFFFF (65536, precision of 0.00137 of degree) */

#define TABLESIZE	1024

static double	TrigTable[TABLESIZE];
static double	factor;
static double	rad2table;
static double	deg2table;

void initTrigTable ( void ) {
	double	d;
	int	i;
	
	factor = 1/(TABLESIZE-1);
	for ( i = 0, d = 0.0; i < TABLESIZE; d += PIOVER2*factor, i++ )
		TrigTable[i] = sin( d );
	/* make sure sin(pi/2) = 1 */
	TrigTable[TABLESIZE-1] = 1.0;
	
	rad2table = TWOOVERPI*(TABLESIZE-1);
	deg2table = (TABLESIZE-1)/90;
}

double sinus ( double rad ) {
begin:
	if ( rad <= PIOVER2 )
		return TrigTable[ round(rad*rad2table) ];
	else if ( rad <= PI )
		return TrigTable[ round((TABLESIZE-1)-(rad-PIOVER2)*rad2table) ];
	else if ( rad <= PI3OVER2 )
		return -TrigTable[ round((rad-PI)*rad2table) ];
	else if ( rad <= PI2 )
		return -TrigTable[ round((TABLESIZE-1)-(rad-PI3OVER2)*rad2table) ];
	else {
		/* saves an if in every rad < 2PI */
		rad = fmod( rad, PI2 );
		goto begin;
	}
}

double sinusd ( double deg ) {
	/* this can be improved using linear interpolation!! */
begin:
	if ( deg <= 90 )
		return TrigTable[ round(deg*deg2table) ];
	else if ( deg <= 180 )
		return TrigTable[ round((TABLESIZE-1)-(deg-90)*deg2table) ];
	else if ( deg <= 270 )
		return -TrigTable[ round((deg-180)*deg2table) ];
	else if ( deg < 360 )
		return -TrigTable[ round((TABLESIZE-1)-(deg-270)*deg2table) ];
	else {
		deg = fmod( deg, 360 );
		goto begin;
	}
}

inline double cosinus ( double rad ) {
	return sinus( rad > PI3OVER2 ? rad - PI3OVER2 : rad + PIOVER2 );
}

inline double cosinusd ( double deg ) {
	return sinusd( deg > 270 ? deg - 270 : deg + 90 );
}
