/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/

#include "ulengine.h"
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#define MAXSIZE		512
#define GEOMERROR	"Wrong geometry: file either damaged or no recognized"
#define FATAL		"File fatal error"
#define COMMENT		'#'
#define MEMORY		"Could not allocate memory"

/* read a string for a file, testing for comments and empty lines */
sInt readString ( FILE *fp, uByte *buffer, sInt *size ) {
	uByte		*t;
	sInt		lines = 0;
	
	if ( feof(fp) )
		return 0;
	do {
#ifdef USE_GETLINE
		getline( buffer, size, fp );
#else
		fgets( buffer, MAXSIZE, fp );
#endif
		lines++;
		if ( (t = strchr( buffer, COMMENT )) )
			*t = '\0';
		t = buffer + strlen(buffer);
		while ( isspace( *(--t) && t >= buffer ) )
			*t = '\0';
	} while ( buffer[0] == '\0' && !feof(fp) );
	return lines;
}

/* Read geometry from a filename and store in d. The doodad is passed as
argument because you may already have allocated it. No checks are done
to see if parts of it are already allocated! */
sInt readGeometry ( uByte *filename, DOODAD_UL *d ) {
	sInt 		i, j, k, n, npoints, lines = 0, color[3], size = 0;
#ifdef USE_GETLINE
	uByte 		*t, *buffer;
#else
	uByte 		*t, buffer[MAXSIZE];
#endif
	Bool 		centroidb = True, sphere = True;
	FILE 		*fp;
	POINT_UL 	*p;
	VECTOR_UL 	p1, p2;
#ifndef CLIENT
	Bool		onecolor;
#endif
	
	if ( !d ) { /* allocate  memory */
		if ( !(d = (DOODAD_UL *)malloc(sizeof(DOODAD_UL))) ) {
			fprintf( stderr, "\n%s: %s for doodad %s, readGeom\n",
				FATAL, MEMORY, filename );
			return FATALERROR;
		}
	}
	
	if ( !(fp = fopen ( filename, "r" )) ) {
		fprintf( stderr, "\n%s: readGeom, Could not open %s", FATAL,
				filename );
		return FATALERROR;
	}

	/* check for 'uglf' header */
	lines += readString( fp, buffer, &size );
	if ( !strstr( buffer, "ulgf" ) ) {
		fprintf( stderr, "\n%s: %s\n \"ulgf\" not found, %s", FATAL,
				GEOMERROR, filename );
#ifdef USE_GETLINE
	free( buffer );
#endif
		return FATALERROR;
	}

	/* read version */
	lines += readString( fp, buffer, &size );
/* fix this 
	sscanf( buffer, "%s %s", NULL, NULL ); */
	if ( !strstr( buffer, "version" ) ) {
		fprintf( stderr, "\n%s: %s\n \"version\" not found, %s", FATAL,
				GEOMERROR, filename );
#ifdef USE_GETLINE
	free( buffer );
#endif
		return FATALERROR;
	}
/*	if ( version != 1.0 )
		fprintf( stderr, "\n%s: unknown version %f; trying to read anyway",
			GEOMERROR, f ); */

	/* read textures */
	lines += readString( fp, buffer, &size );
	if ( (t = strstr( buffer, "textures" )) != NULL ) {
		t += strlen("textures");
/***** This whole thing should be corrected */
		sscanf( t, "%s %d", NULL, &n );
		for ( i = 0; i < n; i++ ) {
			fscanf( fp, "%d:%s", &j, t );
#ifdef CLIENT
			/* open texture here, and save in texture[j] */
#endif
		}
		lines += readString( fp, buffer, &size );
	}

	/* object name */	
	d->name = strdup( buffer );
	
	/* reading precalculated options */
	lines += readString( fp, buffer, &size );
	t = buffer;
	while ( isspace( *t ) )	t++;
	if ( *t != ':' ) { /* centroid */
		if ( sscanf( t, "%lf %lf %lf", &(d->centroid.x), &(d->centroid.y),
				&(d->centroid.z)) != 3 ) {
			fprintf( stderr, "%s: %s line %d: Center of mass\n",
					GEOMERROR, filename, lines );
			zeroV( &d->centroid );
		}
		else centroidb = False;
	}
	if ( *(++t) != ':' ) { /* bounding sphere */
		if ( sscanf( t, "%lf %lf %lf", &(d->sphere.center.x), 
			  &(d->sphere.center.y), &(d->sphere.center.z)) != 3 ) {
			fprintf( stderr, "%s: %s line %d:Sphere center\n",
					GEOMERROR, filename, lines );
			zeroV( &(d->sphere.center) );
		}
		else sphere = False;
		while ( isspace( *t ) )	t++;
		if ( *t == ';' ) {
			if ( sscanf( t, "%lf", &(d->sphere.radius)) != 1 ) {
				fprintf( stderr, "%s: %s line %d:Sphere radius\n",
					GEOMERROR, filename, lines );
				d->sphere.radius = 0;
			}
			else sphere = False;
		}
		else {
			fprintf( stderr, "%s: %s line %d:Sphere radius\n",
					GEOMERROR, filename, lines );
			d->sphere.radius = 0;
		}
	}

	/* read number of points */
	lines += readString( fp, buffer, &size );
	if ( strstr( buffer, "vertices" ) ) {
		t += strlen("vertices");
		sscanf( t, "%d", &(d->nvert) );
		lines += readString( fp, buffer, &size );
	}
	
	p = (POINT_UL *) malloc (d->nvert*sizeof(POINT_UL));
	if ( !p ) {
		fprintf( stderr, "%s for %d points: readGeom:%s\n", MEMORY,
				d->nvert, filename );
#ifdef USE_GETLINE
	free( buffer );
#endif
		return FATALERROR;
	}
	for ( i = 0; i < d->nvert; i++ ) {
		p[i].vect = (VECTOR_UL *) malloc ( sizeof(VECTOR_UL *) );
		if ( !p[i].vect ) {
			fprintf( stderr, "%s for %d vectors, readGeom:%s\n", MEMORY,
					d->nvert, filename );
			for ( j = 0; j < i-1; j++ )
				free( p[j].vect );
			free( p );
#ifdef USE_GETLINE
			free( buffer );
#endif
			return FATALERROR;
		}
	}
	
	/* read vertices */
	for ( i = 0; i < d->nvert; i++ ) {
		lines += readString( fp, buffer, &size );
		t = buffer;
		sscanf( t, "%d%n", &j, &n );
		t += n;
		while ( isspace( *t ) )	t++;
		k = sscanf( t, "%lf %lf %lf%n", &(p[j].vect->x), &(p[j].vect->y), 
				&(p[j].vect->z), &n );
		if ( k != 3 && k != 4 ) { /* sscanf return value is not defined
						when using %n, according to 
						scanf() man page */
			fprintf( stderr, "%s, %s: %s line %d:Sphere radius\n",
					FATAL, GEOMERROR, filename, lines );
			for ( i = 0; i < d->nvert; i++ ) {
				if ( p[i].norm )
					free( p[i].norm );
				free( p[i].vect );
			}
			free( p );
#ifdef USE_GETLINE
			free( buffer );
#endif
			return FATALERROR;
		}
		t += n;
		while ( isspace( *t ) ) t++;
		if ( *t == 'n' ) {
			p[j].norm = (VECTOR_UL *)malloc(sizeof(VECTOR_UL));
			if ( !p[j].norm ) {
				fprintf( stderr, "%s norm, readGeom:%s\n", MEMORY,
						filename );
				for ( i = 0; i < d->nvert; i++ ) {
					free( p[i].vect );
					if ( p[i].norm )
						free( p[i].vect );
				}
				free( p );
#ifdef USE_GETLINE
				free( buffer );
#endif
				return FATALERROR;
			}
			sscanf( t, "%lf %lf %lf", &p[j].norm->x, &p[j].norm->y,
					&p[j].norm->z );
		}
	}

	/* check for eov */
	lines += readString( fp, buffer, &size );
	if ( !strstr( buffer, "eov" ) ) {
		fprintf( stderr, "%s: %s\n \"eov\" is not where it should be, line %s\n",
				FATAL, GEOMERROR, filename );
		for ( i = 0; i < d->nvert; i++ ) {
			free( p[i].vect );
			if ( p[i].norm )	free( p[i].vect );
		}
		free( p );
#ifdef USE_GETLINE
	free( buffer );
#endif
		return FATALERROR;
	}
	
	/* read polygons */
	lines += readString( fp, buffer, &size );
	sscanf( buffer, "%d", &(d->npoly) );
	d->polygons = (POLY_UL **) malloc (d->npoly*sizeof(POLY_UL *));
	/* read polys */
	for ( i = 0; i < d->npoly; i++ ) {
		lines += readString( fp, buffer, &size );
		t = buffer;
		sscanf( buffer, "%d%n", &npoints, &n );
		t += n;
		if ( strstr( t, "onecolor" ) ) {
			t += strlen( "onecolor" );
			j = 1;
#ifdef CLIENT
			d->polygons[i]->onecolor = True;
#else
			onecolor = True;
#endif
			sscanf( t, "%d,%d,%d%n", &color[0], &color[1], 
					&color[2], &n );
			t += n;
		}
		else {
			j = npoints;
#ifdef CLIENT
			d->polygons[i]->onecolor = False;
#else
			onecolor = False;
#endif
		}
		while ( isspace( *t ) )	t++;
		if ( !(d->polygons[i]->points = (POINT_UL *) malloc (npoints*sizeof(POINT_UL)))
#ifdef CLIENT
		|| !(d->polygons[i]->color = (ggi_color *) malloc (j*sizeof(ggi_color)) )) {
#else
		) {			
#endif
			
			fprintf( stderr, "%s: %s, points[%d]",
				FATAL, MEMORY, i );
			for ( i = 0; i < d->nvert; i++ ) {
				free( p[i].vect );
				if ( p[i].norm )	free( p[i].vect );
			}
			free( p );
#ifdef USE_GETLINE
			free( buffer );
#endif
			return FATALERROR;
		}
		
#ifdef CLIENT	/* one color? */
		if ( d->polygons[i]->onecolor == True )
			d->polygons[i]->color[0] = RGBAtoColor( color[0], color[1], 
					color[2], 0 );
#endif
		/* read poly */
		for ( j = 0; j < npoints; j++ ) {
			sscanf( buffer, "%d%n", &k, &n );
			d->polygons[i]->points[j].vect = p[k].vect;
			if ( p[k].norm )
				d->polygons[i]->points[j].norm = p[k].norm;
			t += n;
			while ( isspace( *t ) )	t++;
			
			/* (u, v) coordinates */
			if ( *t == '(' ) {
#ifdef CLIENT
				sscanf( t, "(%f %f)%n", 
					&d->polygons[i]->points[j].u,
					&d->polygons[i]->points[j].v, &n );
#else
				sscanf( t, "(%f %f)%n", NULL, NULL, &n );
#endif
				t += n;
			}
			while ( isspace( *t ) )	t++;
			
			/* read color */
#ifdef CLIENT
			if ( d->polygons[i]->onecolor == False ) { 
#else
			if ( onecolor == False ) { 
#endif
				sscanf( t, "%d,%d,%d%n", &color[0], &color[1], 
					&color[2], &n );
#ifdef CLIENT
				d->polygons[i]->color[j] = RGBAtoColor( color[0], color[1],
					color[2], 0 );
#endif
				t += n;
			}
		}
		
		pointstoV( &d->polygons[i]->points[1], &d->polygons[i]->points[0], &p1 );
		pointstoV( &d->polygons[i]->points[2], &d->polygons[i]->points[0], &p2 );
		crossprodV( &p2, &p1, &(d->polygons[i]->normal) );
		normalizeV( &(d->polygons[i]->normal) );
		centerP( d->polygons[i], &(d->polygons[i]->center) );		
		d->polygons[i]->visible = True;
#ifdef CLIENT
		/* now texture */
		if ( *t == 't' ) {
			sscanf( t, "%d", &k );
			/** handle texture */
		}
#endif
	}
	
	/* calculate centroid and sphere if needed */
	if ( sphere == True ) 
		d->sphere = circSphere( d );
	if ( centroidb == True ) 
		centroid( d, &(d->centroid) );

	free( p ); /* but not the vertices or normals, of course */
#ifdef USE_GETLINE
	free( buffer );
#endif
	fclose( fp );
	
	return OK;
}
