/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#ifdef CLIENT
#include "ulengine.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define VFACTOR		1 	/* used in frustrum... is 1 right? */
#define PROJFACTOR	1 	/* usually one... */

#define ZNEARDIST	0.1	/* default distances from znear and zfar */
#define ZFARDIST	1000	/* planes to camera, in meters */

uByte		*basis;
sInt		framenum;
FRAME_UL	frame;

static CAMERA_UL *camera;	/* this is a pointer to the camera */
static sInt 	swidth_2, sheight_2, swidth, sheight;
static POLY_UL	*presult;
static VECTOR_UL camposl, camdirl;	/* camera position and direction in
					  local coordinates */
static BASISCHANGE_UL localtocam;

/* set window width */
static void changeHW ( sInt width, sInt height ) {
	swidth = width;
	sheight = height;
	swidth_2 = width>>1;
	sheight_2 = height>>1;
}

#define PRESULT_POINTS	30		/* high enough, I guess... Even Gauss
					   could make his 17-gon! */

/* allocate memory for a temporary polygon */
sInt initProject ( sInt width, sInt height ) {
	sInt			i, j;

	/* allocate zbuffer */
	if ( !(frame.zbuf = (fFloat **)malloc(mode.visible.y*sizeof(fFloat *))) ) {
		fprintf( stderr, "Couldn't allocate zbuffer 1\n" );
		return FATALERROR;
	}
	/* FIRST Y, then X. This is to optimize drawHLine */
	for ( i = 0; i < mode.visible.y; i++ ) {
		if ( !(frame.zbuf[i] = (fFloat *)malloc(mode.visible.x*sizeof(fFloat))) ) {
			fprintf( stderr, "Couldn't allocate zbuffer 2\n" );
			return FATALERROR;
		}
		for ( j = 0; j < mode.visible.x; j++ ) {
		      frame.zbuf[i][j] = 1.0;
		}
	}
	/* allocate framebuffer */
	if ( !(frame.color = (ggi_color **)malloc(mode.visible.y*sizeof(ggi_color *))) ) {
		fprintf( stderr, "Couldn't allocate color frame 1\n" );
		return FATALERROR;
	}
	/* FIRST Y, then X. This is to optimize drawHLine */
	for ( i = 0; i < mode.visible.y; i++ ) {
		if ( !(frame.color[i] = (ggi_color *)malloc(mode.visible.x*sizeof(ggi_color))) ) {
			fprintf( stderr, "Couldn't allocate color frame 2\n" );
			return FATALERROR;
		}
	}
	/* temporary buffer for output */
	if ( !(frame.buffer = (uByte *)malloc(mode.visible.x*sizeof(uLong))) ) {
		fprintf( stderr, "Couldn't allocate tmp buffer\n" );
		return FATALERROR;
	}

#ifdef CHRIS_HO_CODE
	/* allocate zbuffer */
	if (!(zbuf = (PIXEL_UL **) malloc(mode.visible.y * sizeof(PIXEL_UL *)))) {
	   fprintf(stderr, "Couldn't allocate zbuffer 1\n");
	   return FATALERROR;
	}

	/* FIRST Y, then X. This is to optimize drawHLine */
	for (i = 0; i < mode.visible.y; i++)
	{
	   if (!(zbuf[i] = (PIXEL_UL *) calloc(mode.visible.x, sizeof(PIXEL_UL))))
	   {
	      fprintf(stderr, "Couldn't allocate zbuffer 2\n");
	      return FATALERROR;
	   }
	}

	  /*
	   * Since calloc() sets all values to 0, z buffering won't work properly.
	   * We need to set the z depth to 1.0, and set the color to SOMETHING.
	   */

	for (i = 0; i < mode.visible.y; i++)
	{
	   for (j = 0; j < mode.visible.x; j++)
	   {
	      /* NULL means use the default background color. */
	      zbuf[i][j].color = NULL;
	      zbuf[i][j].depth = 1.0;
	   }
	}
#endif

	if ( !(presult = (POLY_UL *) malloc (sizeof(POLY_UL))) ) {
		fprintf( stderr, "initProject, 1\n" );
		return FATALERROR;
	}
	presult->npoints = PRESULT_POINTS;
	if ( !(presult->points = (POINT_UL *)malloc(presult->npoints*sizeof(POINT_UL)))) {
		fprintf( stderr, "initProject, 2\n" );
		return FATALERROR;
	}
	for ( i = 0; i < presult->npoints; i++ ) {
		if ( !(presult->points[i].vect = (VECTOR_UL *)malloc(sizeof(VECTOR_UL)))) {
			fprintf( stderr, "initProject, 2\n" );
			return FATALERROR;
		}
		if ( !(presult->points[i].norm = (VECTOR_UL *)malloc(sizeof(VECTOR_UL)))) {
			fprintf( stderr, "initProject, 3\n" );
			return FATALERROR;
		}
	}
	if ( !(presult->color = (ggi_color *)malloc( presult->npoints
			* sizeof(ggi_color))) ) {
		fprintf( stderr, "initProject, 4\n" );
		return FATALERROR;
	}
	if ( !(presult->flatcolor = (ggi_color *)malloc(sizeof(ggi_color))) ) {
		fprintf( stderr, "initProject, 5\n" );
		return FATALERROR;
	}
	presult->visible = True;
	changeHW( width, height );
	return OK;
}

/* frees memory */
void closeProject ( void ) {
	sInt 			i;

	free( presult->color );
	free( presult->flatcolor );
	for ( i = 0; i < presult->npoints; i++ ) {
		free( presult->points[i].vect );
		free( presult->points[i].norm );
	}
	free( presult->points );
	free( presult );

	/* bye zbuffer... */
	for ( i = 0; i < mode.visible.y; i++ ) {
		free( frame.zbuf[i] );
		free( frame.color[i] );
	}
	free( frame.zbuf );
	free( frame.color );
}

/* set cam as the rendering camera */
void chooseCamera ( CAMERA_UL *cam ) {
	camera = cam;
	setCameraFrustrum();
}

/* this function is adapted from code found on the net */
void setCameraFrustrum ( void ) {
	dFloat			ah, av, sh, sv, ch, cv;

	/* these are all in camera coordinates */
	ah = atan2( swidth_2, VFACTOR )-0.00001;
	av = atan2( sheight_2, VFACTOR )-0.00001;
	sh = sin( ah );
	sv = sin( av );
	ch = cos( ah );
	cv = cos( av );
	
	/* left */
	initV( &(camera->frustrum.side[0].normal), ch, 0.0, sh );
	camera->frustrum.side[0].dist = 0;
	normalizeV( &(camera->frustrum.side[0].normal) );
	/* bottom */
	initV( &(camera->frustrum.side[1].normal), 0.0, -cv, sv );
	camera->frustrum.side[1].dist = 0;
	normalizeV( &(camera->frustrum.side[1].normal) );
	/* right */
	initV( &(camera->frustrum.side[2].normal), -ch, 0.0, sh );
	camera->frustrum.side[2].dist = 0;
	normalizeV( &(camera->frustrum.side[2].normal) );
	/* top */
	initV( &(camera->frustrum.side[3].normal), 0.0, cv, sv );
	camera->frustrum.side[3].dist = 0;
	normalizeV( &(camera->frustrum.side[3].normal) );
	/* z-near clipping plane */
	initV( &(camera->frustrum.znear.normal), 0.0, 0.0, 1.0 );
	camera->frustrum.znear.dist = -ZNEARDIST;
	normalizeV( &(camera->frustrum.znear.normal) );
	/* z-far clipping plane */
	initV( &(camera->frustrum.zfar.normal), 0.0, 0.0, 1.0 );
	camera->frustrum.zfar.dist = ZFARDIST;
	normalizeV( &(camera->frustrum.zfar.normal) );
	
	basisVA( &camera->dir, camera->angle, camera->worldtocam );
	invertbasis( camera->worldtocam, camera->camtoworld );
}

void setCameraV ( CAMERA_UL *cam, VECTOR_UL *wpos, VECTOR_UL *dir, dFloat angle ) {
	if ( wpos != NULL ) {
		copyV( &cam->pos, wpos );
	}
	if ( dir != NULL ) {
		copyV( &cam->dir, dir );
		normalizeV( &cam->dir );
	}
	cam->angle = angle;
}
		
void setCameraD ( CAMERA_UL *cam, dFloat wposx, dFloat wposy, dFloat wposz,
		dFloat dirx, dFloat diry, dFloat dirz, dFloat angle ) {
	cam->pos.x = wposx;		cam->pos.y = wposy;
	cam->pos.z = wposz;		cam->dir.x = dirx;
	cam->dir.y = diry;		cam->dir.z = dirz;
	normalizeV( &cam->dir );
	cam->angle = angle;
}

/* passing 0 turns zfar off */
void setCameraZ ( CAMERA_UL *cam, dFloat znear, dFloat zfar ) {
	if ( znear <= 0 ) /* znear == 0 means everything projected in a point */
		cam->znear = ZNEARDIST;
	else
		cam->znear = znear;
	if ( zfar < znear && zfar != 0 )
		cam->zfar = ZFARDIST;
	else
		cam->zfar = zfar;
}

/**
 * This function transforms a polygon from world coordinates to camera
 * coordinates. This transformation is the first step in drawing the
 * polygon on the screen.
 *
 * @param poly a polygon to transform
 * @param clip do we clip or not
 * @return the transformed polygon
 * @author Chris Ho
 * @version 99.08.04
 **********************************************************************/
POLY_UL *worldToCameraP ( POLY_UL *poly, Bool clip ) {
	MATRIX_UL 		m;
	VECTOR_UL		right, forward, up, position, upHint = {0, 0, 1};
	sInt			i;

	copyV( &forward, &camera->dir );
	/* camera->dir is supposed to be normalized */

	/*
	 * For right handed camera coordinates, use:
	 * crossprodV(&upHint, &forward, &left);
	 */
	crossprodV( &forward, &upHint, &right );
	normalizeV( &right );
	/* Actually, up is down. */
	crossprodV( &forward, &right, &up );
	scalarMV( &camera->pos, -1, &position );

	/*
	 * OPTIMIZATION: Store the matrix in the camera structure, recompute
	 * when it changes
	 */

	m[0][0] = right.x;   m[0][1] = up.x;   m[0][2] = forward.x;   m[0][3] = 0;
	m[1][0] = right.y;   m[1][1] = up.y;   m[1][2] = forward.y;   m[1][3] = 0;
	m[2][0] = right.z;   m[2][1] = up.z;   m[2][2] = forward.z;   m[2][3] = 0;
	m[3][0] = dotprodV( &position, &right );
	m[3][1] = dotprodV( &position, &up );
	m[3][2] = dotprodV( &position, &forward );
	m[3][3] = 1;

	for (i = 0; i < poly->npoints; i++) {
		multVM(m, poly->points[i].vect, presult->points[i].vect);
		/* z is done */
	}

	return presult;
}

/**
 * Convert a polygon from camera coodinates into screen coordinates.
 *
 * @param poly the polygon to move to screen space
 * @return the polygon transformed into screen space
 * @author Chris Ho
 * @version 99.08.04
 **********************************************************************/
POLY_UL *cameraToScreenP ( POLY_UL *poly ) {
	dFloat			zDepth, zHeight, z;
	sInt			i;
	MATRIX_UL		perspectiveTransform;

	/*
	 * OPTIMIZATION: store the matrix in the camera structure. Only
	 * recompute if the camera changes.
	 */
	zDepth = camera->zfar - camera->znear;
	zHeight = camera->znear / 1; /* Replace 1 with size of view plane. */

	perspectiveTransform[0][0] = zHeight;
	perspectiveTransform[0][1] = 0;
	perspectiveTransform[0][2] = 0;
	perspectiveTransform[0][3] = 0;

	perspectiveTransform[1][0] = 0;
	perspectiveTransform[1][1] = zHeight;
	perspectiveTransform[1][2] = 0;
	perspectiveTransform[1][3] = 0;

	perspectiveTransform[2][0] = 0;
	perspectiveTransform[2][1] = 0;
	perspectiveTransform[2][2] = camera->zfar / zDepth;
	perspectiveTransform[2][3] = 1;

	perspectiveTransform[3][0] = 0;
	perspectiveTransform[3][1] = 0;
	perspectiveTransform[3][2] = -camera->zfar * camera->znear / zDepth;
	perspectiveTransform[0][1] = 0;

	for ( i = 0; i < poly->npoints; i++ ) {
		z = poly->points[i].vect->z;
		multVM( perspectiveTransform, poly->points[i].vect,
		       presult->points[i].vect );
		/*
		 * We must divide the result by the z in camera space because the
		 * matrix multiplication only works on R3 coordinates. Dividing by
		 * z scales the image to its depth.
		 */
		presult->points[i].vect->x /= z;
		presult->points[i].vect->y /= z;
		presult->points[i].vect->z /= z;
	}
	return presult;
}

/**
 * Put the polygon in the z buffer. We must check each z buffer point
 * to see if the corresponding polygon point is in front. If so, replace
 * the color with the polygon's color.
 *
 * @param poly the polygon to draw on the z buffer
 * @author Chris Ho
 * @version 99.08.04
 **********************************************************************/
void screenToZBufP ( POLY_UL *poly ) {
	sInt 			x, y;
	sInt 			i;

	for ( i = 0; i < poly->npoints; i++ ) {
		x = poly->points[i].vect->x;
		y = poly->points[i].vect->y;

		copyColor( &frame.color[x][y], poly->flatcolor );
	}
}

/* project a doodad */
/**/void projectD ( DOODAD_UL *doodad ) {
	VECTOR_UL 	v;
	Bool 		clip = False;
	dFloat 		d;
	sInt		i;
	POLY_UL		*cameraPoly, *screenPoly;

#ifdef BROKEN /* sphere function is broken */
	/* test of the sphere */
	changebasis( &(doodad->world), camera->worldtocam, &v );
	/* we must have the distance from v to plane < doodad->radius */
	for ( i = 0; i < 4; i++ ) {
		/* distance from v to plane.
		   we should do subV( &v, pointofplane, &v ), but
		   all the planes contain (0,0,0)! */
		d = dotprodV( &v, &(camera->frustrum.side[i].normal) );
		/* it's also not necessary to divide by the modulus of the
		   normal, as it's already normalized */
		
		if ( d < -doodad->sphere.radius ) 
			/* which also means d < 0, so 'out of plane' */
			return;
		else if ( d > -doodad->sphere.radius && d < doodad->sphere.radius )
			/* i.e., part of the sphere is in, part is out, so
			   we'll check for clipping */
			clip = True;
	}

	/* now the znear & zfar planes */
	if ( camera->znear ) {
		if ( dotprodV( &v, &(camera->frustrum.znear.normal) ) < 
				-doodad->sphere.radius )
			return;
		else if ( d > -doodad->sphere.radius && d < doodad->sphere.radius )
			clip = True;
	}

	if ( camera->zfar ) {
		if ( dotprodV( &v, &(camera->frustrum.zfar.normal) ) < 
				-doodad->sphere.radius )
			return;
		else if ( d > -doodad->sphere.radius && d < doodad->sphere.radius )
			clip = True;
	}
#endif
	
	/*
	multM3( doodad->localtoworld, camera->worldtocam, localtocam );
	*/

	for ( i = 0; i < doodad->npoly; i++ ) {
		/* is the face visible? */
	/*
		if ( dotprodV( &(doodad->polygons[i]->normal), &camdirl ) > 0 )
	*/
		cameraPoly = worldToCameraP( doodad->polygons[i], clip );
		screenPoly = cameraToScreenP( cameraPoly );
//		screenToZBufP( screenPoly );
		drawPoly( screenPoly );
	}
}

/**
 * Draw all the doodads onto the screen.
 *
 * @param cam the camera to use for drawing, or NULL for the old camera
 * @return OK if we succeeded, FATALERROR on error
 * @author Chris Ho
 * @version 99.08.04
 **********************************************************************/
sInt render (CAMERA_UL *cam) {
	/* A list of doodads to render? Maybe we should pass this in. */
	DOODAD_UL 		*worldDoodad = firstdoodad;

	if ( cam == NULL && camera == NULL ) {
		fprintf( stderr, "No camera!\n" );
		return FATALERROR;
	}

	else if ( camera != cam ) {
		camera = cam;
		setCameraFrustrum();
	}

	while ( worldDoodad ) {
		projectD( worldDoodad );
		worldDoodad = worldDoodad->next;
	}

	return putImage();
}

/* put image on the screen *//* put image on the screen */
sInt putImage ( void ) {
	sInt			i;

	for ( i = 0; i < mode.visible.y; i++ ) {
		ggiPackColors( vis, frame.buffer, frame.color[i], mode.visible.x );
		ggiPutHLine( vis, 0, i, mode.visible.x,	frame.buffer );
		memset( frame.color[i], 0, mode.visible.x*sizeof(ggi_color) );
		memset( frame.zbuf[i], 0, mode.visible.x*sizeof(fFloat) );
	}
	ggiFlush( vis );

	if ( numframes == 2 ) {
		framenum = ( framenum == 0 ? 1 : 0 );
		if ( ggiSetWriteFrame( vis, framenum ) != 0 ||
		     ggiSetDisplayFrame( vis, framenum ) != 0 ) {
			fprintf( stderr, "Cannot set display or write frame\n" );
			return FATALERROR;
		}
	}
	ggiSetGCForeground( vis, black );
	ggiFillscreen( vis );
	if ( db ) {
		basis = im[framenum].fbptr;
	}
	return OK;
}

#endif /* CLIENT */
