/* makedata.c */
/* Placed in public domain by Steve Kirkendall.  Use at your own risk */

/* This is a stand-alone program that creates the "data.h" file which is
 * #included by "sound.c".  "data.h" contains samples for some built-in sounds.
 */

#include <stdlib.h>
#include <math.h>

#define SAMPLES_PER_CYCLE	160 /* for 100Hz at a 16000Hz sample rate */
#define NSAMPLES 		1920

static void maketone(void)
{
	int	i;
	double	r, inc;

	printf("\nstatic short tonedata[%d] = {\n\t", NSAMPLES);
	inc = 2.0 * M_PI / (double)SAMPLES_PER_CYCLE;
	for (i = 0, r = 0.0; i < NSAMPLES; i++, r += inc)
	{
		printf("%d", (short)(32767 * sin(r)));
		if (i == NSAMPLES - 1)
			/* print no whitespace or comma */;
		else if ((i % 10) == 9)
			printf(",\n\t");
		else
			printf(", ");
	}
	printf("\n};\n");
}

static void makebuzz(void)
{
	int	i;

	printf("\nstatic short buzzdata[%d] = {\n\t", NSAMPLES);
	for (i = 0; i < NSAMPLES; i++)
	{
		printf("%d", (i % SAMPLES_PER_CYCLE < SAMPLES_PER_CYCLE/2 ? 32767 : -32767));
		if (i == NSAMPLES - 1)
			/* print no whitespace or comma */;
		else if ((i % 10) == 9)
			printf(",\n\t");
		else
			printf(", ");
	}
	printf("\n};\n");
}

static void makehiss(void)
{
	int	i;

	printf("\nstatic short hissdata[%d] = {\n\t", NSAMPLES);
	for (i = 0; i < NSAMPLES; i++)
	{
		printf("%d", (short)rand());
		if (i == NSAMPLES - 1)
			/* print no whitespace or comma */;
		else if ((i % 10) == 9)
			printf(",\n\t");
		else
			printf(", ");
	}
	printf("\n};\n");
}


static void makebang(void)
{
	int	i;
	double	minusone_to_plusone;

	printf("\nstatic short bangdata[%d] = {\n\t", NSAMPLES);
	for (i = 0; i < NSAMPLES; i++)
	{
		minusone_to_plusone = (double)rand()/RAND_MAX * 2.0 - 1.0;
		printf("%d", (short)((32767.0 * (double)(NSAMPLES - i) * minusone_to_plusone) / (double)NSAMPLES));
		if (i == NSAMPLES - 1)
			/* print no whitespace or comma */;
		else if ((i % 10) == 9)
			printf(",\n\t");
		else
			printf(", ");
	}
	printf("\n};\n");
}


main(void)
{
	printf("/* automatically generated by makedata.c */\n");
	maketone();
	makebuzz();
	makehiss();
	makebang();
	exit(0);
}
