/* unixsnd.h */
/* Placed in public domain by Steve Kirkendall.  Use at your own risk */

#include <stdio.h>

typedef struct sound_s
{
	struct sound_s	*next;		/* pointer to another sound */
	char		*name;		/* name of the sound */
	int		frequency;	/* samples per second */
	int		stereo;		/* 0=mono, 1=interleaved stereo */
	int		nsamples;	/* number of samples */
	short		*sample;	/* sample data */
} sound_t;

typedef struct loop_s
{
	struct loop_s	*next;		/* pointer to another loop */
	char		*name;		/* name of loop, or NULL to play once */
	char		*voice;		/* name of voice, or NULL for none */
	sound_t		*sound;		/* the sound to play */
	char		*later;		/*request to interpret when sound ends*/
	long		pos;		/* position within sound->sample[] */
	long		end;		/* maximum value of pos */
	long		rvbend;		/* maximum value of pos for reverb */
	int		incr;		/* increment, derived from pitch */
	int		left, right;	/* volumes */
	int		pitch;		/* pitch */
} loop_t;

typedef struct
{
	loop_t	*loop;		/* loop to affect */
	sound_t	*sound;		/* new sound for loop */
	char	*later;		/* request to interpret when sound ends */
	int	left, right;	/* new volumes for loop */
	int	pitch;		/* new pitch for loop */
	int	reverb;		/* new reverb for loop */
} cmd_t;

typedef struct
{
	long	left, right;	/* sample totals */
} mixed_t;


extern int	verbose;
extern int	mute;
extern int	output_frequency;

/* extern declarations of functions */
loop_t *loopfind(char *name);
int loopidle(void);
void loopfree(loop_t *loop);
int loopmix(mixed_t *mix, int count);
void loopcmd(cmd_t *cmd);
void loopmute(void);
cmd_t *cmdread(FILE *fp);
cmd_t *cmdparse(char *cmdline);
void cmddefine(const char *name, int namelen, const char *text);
sound_t *soundfind(char *basename);
void soundpath(char *addpath);
int devbits(char *name);
int devstereo(char *name);
int devfrequency(char *name, int stereo, int bits);
int devopen(char *name, int stereo, int bits, int frequency);
void devflush(int fd);
void reverbinit(char *value);
int reverbsilence(mixed_t *buf, int maxsamples);
void reverbsamples(mixed_t *buf, int nsamples);
int reverbidle(void);
